
/*

% The  old way: when the user solves the constraints

compute_proof_bounds(Pred1, Pred2, Lo, Hi) :-
	Pred1 = P/M, Pred2 = Q/N,
	nl, writeln(P/M), nl, write_all(P,M), 
	nl, writeln(Q/N), nl, write_all(Q,N),
	writeln('Give lo,hi proof bounds'),
	nl, read(Lo), read(Hi).

*/

% The new way : Using CLP for constraint solving to obtain proof bounds

:- [clpr].
:- import dump/3 from dump.
:- import append/3, member/2 from basics.

compute_proof_bounds(Pred1, Pred2, Lo, Hi) :-
	compute_lower_bound(Pred1, Pred2, Lo),
	compute_upper_bound(Pred1, Pred2, Hi).

compute_lower_bound(Pred1, Pred2, Lo) :- 
	get_lo_constraints(Pred1, Pred2, LVMap, LC1),
	flatten_conj(LC1, LC),
	( {LC} -> member((X, Pred1, Pred2), LVMap),
		  ( number(X) 
			-> Lo is floor(X)
		        ;  ( var(X) -> dump([X], [x], LCList),
        	                       ( ( LCList = [x =< Lo_float] ; 
		                           LCList = [x < Lo_float] )
			                 -> Lo is floor(Lo_float)
			                 ;  Lo = -inf
                                       )
	                            ;  writeln('ERROR: In Lo Proof bound')
                           )
                  )		            
	       ;  Lo = -inf
        ).

compute_upper_bound(Pred1, Pred2, Hi) :-
	get_hi_constraints(Pred1, Pred2, HVMap, HC1),
	flatten_conj(HC1, HC),
	( {HC} -> member((X, Pred1, Pred2), HVMap),
		  ( number(X) 
			-> Hi is floor(X)+1
		        ;  ( var(X) -> dump([X], [x], HCList),
		                       ( ( HCList = [x >= Hi_float] ;
					   HCList = [x > Hi_float] )
			                  -> Hi is floor(Hi_float)+1
			                  ;  Hi = inf
                                       )
				    ;  writeln('ERROR: In Hi Proof bound')
                           )
                  )
	       ;  Hi = inf
        ).	

%----------------------------------------------------------------

get_hi_constraints(Pred1, Pred2, VMap, HC) :-
	get_hi_ineqs(Pred1, Pred2, [], (X=X), VMap, HC).

get_hi_ineqs(Pred1, Pred2, VmapIn, CIn, VmapOut, COut) :-
	member((_, Pred1,Pred2), VmapIn), !,
	VmapOut = VmapIn, COut = CIn.

get_hi_ineqs(Pred1, Pred2, VmapIn, CIn, VmapOut, COut) :-
	Pred1 == Pred2, !, 
	VmapOut = [(X, Pred1, Pred2)| VmapIn],
	COut = (CIn,(X = 0.0)).

get_hi_ineqs(Pred1, Pred2, VmapIn, CIn, VmapOut, COut) :-
	find_clauses(Pred1, CList1),
	find_clauses(Pred2, CList2),
	axiom_clausemap(CList2, CList1, CList1_mapped),
	VmapIn1 = [(_X,Pred1,Pred2)|VmapIn],
	clause_hi_ineqs(CList1_mapped, CList2, VmapIn1, CIn, VmapOut, COut).

clause_hi_ineqs(_, [], V, C, V, C).
clause_hi_ineqs([C|Rest1], [D|Rest2], VIn, CIn, VOut, COut) :-
	clause_hi_ineq(C, D, VIn, CIn, V1, C1),
	clause_hi_ineqs(Rest1, Rest2, V1, C1, VOut, COut).

clause_hi_ineq(C, D, VIn, CIn, VOut, COut) :-
	get_clause(C, (Hd :- B), (_, Gamma)),
	get_clause(D, (Hd1 :- B1), (Mu, _)),
	exec_builtins(B, Bd), exec_builtins(B1, Bd1),
	body_hi_ineqs(Bd, Bd1, VIn, CIn, VOut, Constraints),
	construct_hi_ineq(Hd, Hd1, Bd, Bd1, Gamma, Mu, VOut, Ineq),
	COut = (Constraints, Ineq).

construct_hi_ineq(Hd, Hd1, Bd, Bd1, Gamma, Mu, VMap, Inequality) :-
	get_attr_var(VMap, Hd, Hd1, X),
	get_rhs_exp(VMap, Bd, Bd1, Exp),
	Ineq = (X >= Gamma - Mu + Exp),
	simplify_ineq(Ineq, Inequality).


body_hi_ineqs([], [], V, C, V, C).
body_hi_ineqs([A|Rest], [A1|Rest1], VIn, CIn, VOut, COut) :-
	functor(A, P, M), functor(A1, Q, N),
	get_hi_ineqs(P/M, Q/N, VIn, CIn, V1, C1),
	body_hi_ineqs(Rest, Rest1, V1, C1, VOut, COut).

%---------------------------------------------------------------

get_lo_constraints(Pred1, Pred2, VMap, LC) :-
	get_lo_ineqs(Pred1, Pred2, [], (X=X), VMap, LC).


get_lo_ineqs(Pred1, Pred2, VmapIn, CIn, VmapOut, COut) :-
	member((_, Pred1,Pred2), VmapIn),
	!, VmapOut = VmapIn, COut = CIn.

get_lo_ineqs(Pred1, Pred2, VmapIn, CIn, VmapOut, COut) :-
	Pred1 == Pred2, !, 
	VmapOut = [(X, Pred1, Pred2)| VmapIn],
	COut = (CIn,(X = 0)).

get_lo_ineqs(Pred1, Pred2, VmapIn, CIn, VmapOut, COut) :-
	find_clauses(Pred1, CList1),
	find_clauses(Pred2, CList2),
	axiom_clausemap(CList1, CList2, CList2_mapped),
	VmapIn1 = [(_X,Pred1,Pred2)|VmapIn],
	clause_lo_ineqs(CList1, CList2_mapped, VmapIn1, CIn, VmapOut, COut).

axiom_clausemap([], L, L).
axiom_clausemap([C|Rest], CL, [D|CL_mapped]) :-
	get_clause(C, (H :- B), _), 
	member(CL, D, Pre, Post),
	get_clause(D, (Hd :- Bd), _),	
	exec_builtins(B, B1), exec_builtins(Bd, Bd1),
	clause_inst_chk((H :- B1), (Hd :- Bd1)),
	append(Post, Pre, CL_rest), % just an optimization : not reqd.
	axiom_clausemap(Rest, CL_rest, CL_mapped).

clause_lo_ineqs([], _, V, C, V, C).
clause_lo_ineqs([C|Rest1], [D|Rest2], VIn, CIn, VOut, COut) :-
	clause_lo_ineq(C, D, VIn, CIn, V1, C1),
	clause_lo_ineqs(Rest1, Rest2, V1, C1, VOut, COut).


clause_lo_ineq(C, D, VIn, CIn, VOut, COut) :-
	get_clause(C, (Hd :- B), (Gamma, _)),
	get_clause(D, (Hd1 :- B1), (_,Mu)),
	exec_builtins(B, Bd), exec_builtins(B1, Bd1),
	body_lo_ineqs(Bd, Bd1, VIn, CIn, VOut, Constraints),
	construct_lo_ineq(Hd, Hd1, Bd, Bd1, Gamma, Mu, VOut, Ineq),
	COut = (Constraints, Ineq).

body_lo_ineqs([], [], V, C, V, C).
body_lo_ineqs([A|Rest], [A1|Rest1], VIn, CIn, VOut, COut) :-
	functor(A, P, M), functor(A1, Q, N),
	get_lo_ineqs(P/M, Q/N, VIn, CIn, V1, C1),
	body_lo_ineqs(Rest, Rest1, V1, C1, VOut, COut).

construct_lo_ineq(Hd, Hd1, Bd, Bd1, Gamma, Mu, VMap, Inequality) :-
	get_attr_var(VMap, Hd, Hd1, X),
	get_rhs_exp(VMap, Bd, Bd1, Exp),
	Ineq = (X =< Gamma - Mu + Exp),
	simplify_ineq(Ineq, Inequality).

simplify_ineq(I1, I2) :-
	I1 = (X =< Gamma - Mu + E1), !,
	annot_exp(Gamma-Mu, Annot),
	( Annot == inf -> I2 = (X = X) % satisfiable
		       ;  ( Annot == -inf -> I2 = (X > X + 1) % false
					  ;  I2 = (X =< Annot + E1)
                          )
        ).

simplify_ineq(I1, I2) :-
	I1 = (X >= Gamma - Mu + E1), !,
	annot_exp(Gamma-Mu, Annot),
	( Annot == -inf -> I2 = (X = X) % satisfiable
		       ;  ( Annot == inf -> I2 = (X > X + 1) % false
					 ;  I2 = (X >= Annot + E1)
                          )
        ).

          
	

annot_exp(X - Y, -inf) :- X == -inf, Y == inf, !.
annot_exp(X - Y, inf) :- X == inf, Y == -inf, !.
annot_exp(inf - _, inf) :- !.
annot_exp(-inf - _, -inf) :- !.
annot_exp(_ - inf, -inf) :- !.
annot_exp(_ - Y, inf) :- Y == -inf, !.
annot_exp(X - Y, Z) :- integer(X), integer(Y), !, Z is X - Y.
annot_exp(_ - _, _) :- writeln('ERROR: Clause annotations not integers'), fail.



get_rhs_exp(_VMap, [], [], 0).
get_rhs_exp(VMap, [A|Rest], [A1|Rest1], X+Exp) :-
	get_attr_var(VMap, A, A1, X),
	get_rhs_exp(VMap, Rest, Rest1, Exp).


get_attr_var(VMap, A, A1, X) :-
	functor(A, P, M), functor(A1, Q, N),
	member((X, P/M, Q/N), VMap).	


%----------------------------------------------

member([H|T], H, [], T).
member([H|T], X, [H|Pre], Post) :-
	member(T, X, Pre, Post).


%----------------------------------------------

flatten_conj((C1, C2), C) :- 
	!, flatten_conj(C1, F1), flatten_conj(C2, F2),
	append_conj(F1, F2, C).

flatten_conj(C, C).

append_conj((C1, C2), D, (C1, Res)) :- !, append_conj(C2, D, Res).
append_conj(C, D, Res) :- append1(C, D, Res).

append1(C, (D1, D2), (C,D1,D2)) :- !.
append1(C, D, (C, D)).
