
:- import length/2, append/3, member/2 from basics.
:- import variant/2, subsumes/2 from subsumes.


% UNCONDITIONAL FOLDING

% fold(Folded, Folder)
% 	Folded is the set of folded clauses in the current program.
% 	Folder is the set of folder clauses in the original program.
% 	clauses of current program represented as :
% 		clauseI(ClauseId, Clause, Annot)
% 	clauses of original program represented as :
%		clause0(ClauseId, Clause, Annot)

fold(Folded, P/N, NewClause) :-
	!, find_folder_clauses(P/N, Folder),
	fold(Folded, Folder, NewClause).

fold(Folded, Folder, NewClause) :-
	retractall(before_gen(_,_,_)),
	( gen_check(Folded)     % To generalize or not to generalize
		-> fold_gen(Folded, Folder, NewClause)
		;  fold_normal(Folded, Folder, NewClause)
        ).

fold_normal(Folded, Folder, NewClause) :-
 	( foldable(Folded, Folder, C, Biject) 
		-> apply_fold(C, Biject, NewClause)
                ;  fail
        ).

fold_gen(Folded, Folder, NewClause) :-
	( foldable(Folded, Folder, C, Biject) 
		-> apply_fold(C, Biject, NewClause)
                ;  restore_non_gen(Folded), fail
        ).

%--------------------------------------------------------------------

find_folder_clauses(Folder) :-
	get_folder_pred(P/N), find_folder_clauses(P/N, Folder).

get_folder_pred(Pred/Num) :-
	setof(P/N, (clause0(_, (H :- _), _), functor(H,P,N)), List),
	member(Pred/Num, List).

find_folder_clauses(P/N, Folder) :-
	length(L, N), Hd =.. [P|L],
	\+ clause0(_, (Hd :- []), _),  % never have facts as folder
	findall(Id, clause0(Id, (Hd :- _), _), Folder),
        Folder \= [].   % disallow empty folder clause-set

%-----------------------------------------------------------------------

apply_fold(C, Biject, N) :-
	Biject = (Fd, Fr), 
	annot_after_folding(Fd, Fr, Annot),
 	retract_clauses(Fd),
	gen_new_clauseNum(N), 
	assert_clause(N, C, Annot),
	% write_clause(N, C, Annot),  % keep the transformation silent
	update_fold_history(N, Biject).

update_fold_history(N, Biject) :-
	Biject = (Folded, Folder),
	create_listmap(Folded, Folder, ListMap),
	assert(history(N, fold, ListMap)).

annot_after_folding(Fd, Fr, (A, A1)) :-
	minus_all(Fd, Fr, Alist, A1list),
	min_annot(Alist, A), 
	max_annot(A1list, A1).

minus_all([], [], [], []).
minus_all([C|Crest], [D|Drest], [A1|A1rest], [A2|A2rest]) :-
	clauseI(C, _, (Ac1, Ac2)), clause0(D, _, (Ad1, Ad2)),
	minus_annot(Ac1, Ad2, A1), minus_annot(Ac2, Ad1, A2),
	minus_all(Crest, Drest, A1rest, A2rest).


% History of a clause n is the clauses from which n was derived.
%    history( ClauseId, TxfType, 
%             List of (SiteClause, Manipulated Atoms, Used Clause)
%           )
% For the history of clauses created by folding we do not maintain
% the manipulated atoms in the current implementation.

create_listmap([], [], []).
create_listmap([Fd|FdRest], [Fr|FrRest], [(Fd, unused, Fr)|X]) :-
	create_listmap(FdRest, FrRest, X).

%------------------------------------------------------------------------

% foldable(Folded, Folder, C, Biject) 
% ------ Checks pre-conditions of a folding step.
%        Returns C, the new clause obtained by folding
%        and Biject, the bijection between folded and folder clauses.

foldable(Folded, Folder, C, Biject) :-
	permute(Folded, Fd), 
	Biject = (Fd, Folder), 
	inst_all(Fd, Folder, Clist, Blist), 
	variant_all(Clist), variant_all(Blist), 
	Clist = [C|_], Blist = [B|_],
	soundnessCheck(B, Folder), !.

%-------------------------------------------------------------------------
% soundnessCheck(B, Folder)
% Checks that Folder is exactly the set of clauses in the 
% original program that unify with B.
	
soundnessCheck(B, Folder) :-
	findall(CId, (clause0(CId, (H :- _), _), unifies(H, B)),
                ClauseList),
	permute(Folder, ClauseList).

%***************************************************************************

inst_all([], [], [], []).
inst_all([Cd|FdRest], [Cr|FrRest], [C|CRest], [B|Brest]) :-
	inst_one(Cd, Cr, C, B),
	inst_all(FdRest, FrRest, CRest, Brest).

% inst_one(Cd, Cr, New, B)
% ----- Cd is the clause Id of a folded clause
%       Cr is the clause Id of the corresponding folder clause
%       New is the clause generated by folding.
%       B is the folder atom.

inst_one(Cd, Cr, New, Hr) :-
	clauseI(Cd, Clause_fd, (Ad, _)),  % look up the folder and
	clause0(Cr, Clause_fr, (_, Ar)),  % and folded clauses
	copy_term(Clause_fd, (Hd :- Bd)), % Do not spoil the pgm. clauses
	copy_term(Clause_fr, (Hr :- Br)), % make a copy
        find_internal_vars((Hr :- Br), IVars),
	match_folder(Br, Bd, Bd_nf, Bd_f), % which part of folded clause?
	subsumes(Br, Bd_f),                % check for subsumption
	check_internal_inst(IVars, [Hd|Bd_nf]),
	append(Bd_nf, [Hr], NewBody),
	New = (Hd :- NewBody),
	compute_vector(Bd_nf, CV),
	plus_annot(Ad, CV, Ad1), greater_annot(Ad1, Ar).

%-------------------------------------------------------------------------

match_folder([], Bd, Bd, []).

match_folder([A|Rest], Bd, Bd_nf, [X|Bd_f1]) :-
	pred_match_member(A, Bd, X, Bd1),
	match_folder(Rest, Bd1, Bd_nf, Bd_f1).
	
pred_match_member(A, L, X, Rest) :-
	pred_match_member(A, L, [], X, Rest).

pred_match_member(A, [B|Tail], Prev, X, Rest) :-
	functor(A, P, N),
	( ( functor(B, P, N), X = B, append(Prev, Tail, Rest)
	  ) ; 
          ( append(Prev, [B], Prev1), 
  	    pred_match_member(A, Tail, Prev1, X, Rest)
          )
        ).


%--------------------------------------------------------------------
% Lowlevel checks for instantiations of internal (i.e. existential)
% variables of a clause; one of the pre-conditions of folding requires
% these instantiations to be fresh variables.

find_internal_vars((Head :- Body), IVars) :-
	get_vars(Head, HVars),
	get_vars(Body, BVars),
	var_setminus(BVars, HVars, IVars).

check_internal_inst(IVars, Rest_of_Clause) :-
	vars_all(IVars), var_distinct_all(IVars),
	get_vars(Rest_of_Clause, RestVars),
	var_setintersect(IVars, RestVars, []).

%====================================================================

% PLANS FOR IMMEDIATE FUTURE :

% Plan to introduce Generalization (and Equality Introduction)
% as part of folding. Thus, generalization is performed based
% on the need to fold.

% OK, now the PLAN BECOMES CODE.

% If you generalize the folded clauses, and then folding fails,
% you should restore the old program (pgm. before generalization)

restore_non_gen([]).
restore_non_gen([N|Rest]) :-
	retractall(clauseI(N,_,_)),
	before_gen(N, C, A),
	assert(clauseI(N, C, A)),
	restore_non_gen(Rest).


% gen_check(Folded)
% --- Check whether Folded Clauses need generalization.
%     If so, then generalize the folded clauses using the
%     the most general folded clause-head i.e. after 
%     generalization, the head of each clause is a variant
%     of the most general folded clause head.

gen_check(Folded) :-
	find_heads(Folded, FHeads),
	( variant_all(FHeads) % well, no need to generalize.
		-> fail
		;  ( most_gen_atom(FHeads, GHead) 
			-> generalize_clauses(Folded, GHead)
			;  fail
                   )
        ).


generalize_clauses([], _) :- !.
generalize_clauses([N|Rest], G) :-
	!, generalize_clause(N, G), 
	generalize_clauses(Rest, G).

% generalize_clause(N, G)
% --- Perform generalization (and equality introduction) on clause with
%     clause-id N. After generalization, the head of the clause must be
%     a variant of G.

generalize_clause(N, G) :-
	clauseI(N, (H :- B), Annot),
	assert(before_gen(N, (H :- B), Annot)),
	( variant(G, H) 
	       -> true
	       ;  generalize_head(H, G, H1, Eqs),
		  generalize_body(B, Eqs, B0),
		  append(B0, Eqs, B1),
		  retractall(clauseI(N,_,_)),
		  assert(clauseI(N, (H1 :- B1), Annot))
	).  
		   
% generalize_head(H,G,H1,Eqs)
% --- Assumption : H = G\theta (H is an instance of G)
%     Generalize H to H1 which is a variant of G.
%     Return substitutions resulting from the generalization in Eqs

generalize_head(H, G, H1, Eqs) :-
	copy_term(G, H1),        % caution: if G shares a variable with 
			         %          the clause whose head is H
	copy_term(H1,H2,Eqs),    % copying to get explicit instantiations
	H2 = H.

generalize_body([], _, []).
generalize_body([A|Rest], Eqs, [A1|Rest1]) :-
	generalize_atom(A, Eqs, A1),
	generalize_body(Rest, Eqs, Rest1).

generalize_atom(A, _, A) :- builtin(A), !.
generalize_atom(A, Eqs, B) :-
	A =.. [P|Args], 
	generalize_terms(Args, Eqs, Args1),
	B =.. [P|Args1].

generalize_terms([],_,[]).
generalize_terms([T|Rest], Eqs, [T1|Rest1]) :-
	generalize_term(T, Eqs, T1),
	generalize_terms(Rest, Eqs, Rest1).

generalize_term(X, _, X) :- var(X), !.

generalize_term(T, Eqs, X) :-
	member((X = T1), Eqs), T == T1, !.

generalize_term(T, Eqs, T1) :-
	T =.. [F|Tlist],
	generalize_terms(Tlist, Eqs, Tlist1),
	T1 =.. [F|Tlist1].
      
%--------------------------------------------------------------------
