
:- import conset/2, conget/2 from gensym.
:- import append/3, length/2, member/2, memberchk/2, ground/1 from basics.
:- import subsumes/2, variant/2 from subsumes.
:- import str_cat/3 from string.

:- dynamic shared/1.

%===================================================================
% Lowlevel dirty routines.
%===================================================================

depth(X, 1) :- var(X), !.
depth(T, N) :- 
	T =.. [_F|L], depth_list(L, DList), nat_max(DList, M), N is M+1.

depth_list([], []).
depth_list([T|Rest], [N|NRest]):-
	depth(T, N), depth_list(Rest, NRest).


nat_max(L, N) :- nat_max(L, 0, N).

nat_max([], N, N).
nat_max([H|T], M, N) :-
	( H > M -> nat_max(T, H, N) ; nat_max(T, M, N)).


subs_check(G1, G2) :-
	copy_term(G1, X), copy_term(G2, Y), subsumes(X, Y).


% Standard predicate dependence information inference

pred_tc(X, Y) :- 
	abolish_table_pred(pred_tc1/2), pred_tc1(X, Y).

:- table pred_tc1/2.

pred_tc1(X, Y) :- pred_arc(X, Y).
pred_tc1(X, Y) :- pred_tc1(X, Z), pred_arc(Z, Y).

pred_arc(P/N, Q/M) :-
	length(L, N), Head =.. [P|L],
	\+ predicate_property(Head, built_in),
	clauseI(_, (Hd :- Body), _), unifies(Hd, Head),
	member(X, Body), functor(X, Q, M),
	\+ predicate_property(Head, built_in).

%------------------------------------------------------------------
% find_shared_vars(A, B, VList)
% --  returns list of shared variable of A, B in VList
%     could not have used findall/3, shared_var/3 to implement this.

find_shared_vars(X, Y, VL) :- 
	var(X), var(Y), !, ( X == Y -> VL = [X] ; VL = [] ).
find_shared_vars(X, T, VL) :- 
	var(X), !, (contains_var(T, X) -> VL = [X] ; VL = []).
find_shared_vars(T, X, VL) :- 
	var(X), !, (contains_var(T, X) -> VL = [X] ; VL = []).
find_shared_vars(G, G1, VL) :-
	G =.. [_|L], G1 =.. [_|L1], find_shared_vars_list(L, L1, VL).

find_shared_vars_list([], _, []).

find_shared_vars_list([A|Rest], L, VList) :-
	find_shared_vars1(A, L, VL1),
	find_shared_vars_list(Rest, L, VL2),
	append(VL1, VL2, VList).

:- index find_shared_var1/3-2.

find_shared_vars1(_A, [], []).

find_shared_vars1(A, [B|Rest], VList) :-
	find_shared_vars(A, B, VL1),
	find_shared_vars1(A, Rest, VL2),
	append(VL1, VL2, VList).

%-------------------------------------------------------------
%  shared_var(A, B, X)
% ---  returns X, a shared variable of A, B.

shared_var(X, Y, X) :- var(X), var(Y), !, X == Y.
shared_var(X, T, X) :- var(X), !, contains_var(T, X).
shared_var(T, X, X) :- var(X), !, contains_var(T, X).
shared_var(G, G1, X) :-
	G =.. [_|L], G1 =.. [_|L1], 
        shared_var_list(L, L1, X).

shared_var_list(L, L1, X) :-
	member(A, L), member(A1, L1), shared_var(A, A1, X).

contains_var(Y, X) :- var(Y), !, X == Y.
contains_var(T, X) :- T =.. [_|L], member(A, L),
                      contains_var(A, X).

% maxdepth(X, T, N)
% ---  maximum depth of variable X in term T is integer N.

maxdepth(X, Y, N) :- 
	var(Y), !, (X == Y -> N = 0 ; N = no).
maxdepth(_, T, no) :- ground(T), !.
maxdepth(X, T, N) :-
	T =.. [_|L], maxdepth_list(X, L, Max),
        (Max = no -> N = no ; N is 1 + Max).

:- index maxdepth_list/3-2.

maxdepth_list(_, [], no).
maxdepth_list(X, [T|Rest], N) :-
	maxdepth_list(X, Rest, M),
        maxdepth(X, T, M1), max(M, M1, N).

max(no, no, no) :- !.
max(N,  no, N) :- !.
max(no, N, N) :- !.
max(M, M1, N) :- (M > M1 -> N = M ; N = M1).

another_atom(Body, G, G1) :-
	findGinBody(G, Body, Pre, Post),
        ( member(G1, Pre) ; member(G1, Post) ).

%--------------------------------------------------------------------
% Operations on the clause annotations 

zero(0).

min_annot([H|T], M) :- min_annot(T, H, M).

min_annot([], M, M).
min_annot([H|T], A, M) :- 
	greater_annot(A, H) -> min_annot(T, H, M) 
	                    ;  min_annot(T, A, M).

max_annot([H|T], M) :- max_annot(T, H, M).

max_annot([], M, M).
max_annot([H|T], A, M) :- 
	greater_annot(H, A) -> max_annot(T, H, M) 
	                    ;  max_annot(T, A, M).


% computing the sum of the characteristic vectors of a list of atoms
% get_vector/2 returns the characteristic vector of a single predicate

compute_vector([], Zero) :- 
	zero(Zero).
compute_vector([H|T], N) :-
	compute_vector(T, M), 
        get_char_vector(H, V),
        plus_annot(M, V, N).

get_char_vector(H, 1) :- 
	( nonvar(H) -> true
                    ;  writeln('ERROR: in get_char_vector/2'), fail
        ).

annot_type(X) :- var(X), !, fail.
annot_type(inf).
annot_type(-inf).
annot_type(A) :- integer(A).

% does not worry about the pathological case : inf, -inf

plus(A, B, C) :- integer(A), integer(B), !, C is A + B.
plus(A, B, inf) :- (A == inf ; B == inf), !.
plus(A, B, -inf) :- (A == -inf ; B == -inf).

neg(A, NA) :- integer(A), !, NA is 0 - A.
neg(inf, -inf).
neg(-inf, inf).

minus(A, B, C) :- neg(B, NB), plus(A, NB, C).

greater(A, B) :- integer(A), integer(B), !, A > B.
greater(inf, _).
greater(_, -inf).


plus_annot(A, B, C) :-
	((annot_type(A), annot_type(B)) 
		        -> plus(A, B, C)
                        ;  write('ERROR: Cannot add '),
			   write(A),write(','),writeln(B), fail
        ).

minus_annot(A, B, C) :-
	((annot_type(A), annot_type(B)) 
		        -> minus(A, B, C)
                        ;  write('ERROR: Cannot subtract '),
			   write(A),write(','),writeln(B), fail
        ).

greater_annot(A, B) :-
	( (annot_type(A), annot_type(B))
		-> greater(A, B)
		;  writeln('ERROR : Cannot compare'),
		   write(A),write(','),writeln(B), fail
	).


%---------------------------------------------------------------------------

gen_new_nodeNum(N) :-
	conget(node, N1),
	N is N1 + 1,
	conset(node, N).

gen_new_clauseNum(N) :-
	conget(clause, N1),
 	N is N1 + 1,
	conset(clause,N).

reset(Register) :- conset(Register, 0).

get_register(R, N) :- conget(R, N).

get_new_pred(Pred) :-
	conget(preds, N1),
	N is N1 + 1,
	conset(preds, N),
	num_to_string(N, Ns),
	str_cat('new',Ns,Pred).

% some XSB builtins regarding internal representation

num_to_string(N, Ns) :- number_codes(N, L), atom_codes(Ns, L).

%-------------------------------------------------------------------------
% Operations on lists

% the good old permutation program. Generates permutations of lists.

permute([],[]).
permute([H|T], L) :-
	permute(T, T1), insert(H, T1, L).

insert(X, T, [X|T]).
insert(X, [H|T], [H|T1]) :- insert(X, T, T1).

remove(_, [], _) :-  fail.
remove(X, [X|T], T).
remove(X, [H|T], [H|T1]) :- remove(X, T, T1).

remove_all([], L, L).

remove_all([X|Rest], List, New) :-
	remove(X, List, List1), remove_all(Rest, List1, New).

%-------------------------------------------------------------------
% partitioning/splitting of lists.

partition([], [], []).
partition([H|T], [H|L], R) :- partition(T, L, R).
partition([H|T], L, [H|R]) :- partition(T, L, R).


% finding partitions of size N

partition(L, 0, [], L).

partition([H|T], N, L, [H|Rest]) :-
	N > 0, partition(T, N, L, Rest).

partition([H|T], N, [H|L], Rest) :-
	N > 0, M is N - 1, partition(T, M, L, Rest).


find_pos([H|_], 1, H).
find_pos([_|T], N, X) :-
	N > 1, M is N-1, find_pos(T, M, X).


split_list(L, 0, [], L).
split_list([H|T], N, [H|K1], L2) :-
	N > 0, M is N - 1,
	split_list(T, M, K1, L2).




% findGinBody(G,L,Pre,Post)
%      find atom G in vanilla list L ; 
%      Pre (Post) is the list of atoms in L
%      occurring before (after) G
 
findGinBody(G, L, Pre, Post) :-
	find_in_list(G, L, [], Pre, Post).

find_in_list(G, [G|Rest], Pre, Pre, Rest) :- !.
find_in_list(G, [H|Rest], L, Pre, Post) :-
        append(L, [H], L1),
	find_in_list(G, Rest, L1, Pre, Post).

% commaConvert(CL, VL) :
% 	Convert Commalist CL to VanillaList VL

commaConvert(L1, L2) :-
	is_list(L1) -> L2 = L1 ; commaConvert1(L1, L2).

commaConvert1(CL, VL) :- 
	( CL = (G,Rest) -> commaConvert1(Rest,VL1), VL = [G|VL1]
                        ;  ( CL = true -> VL = []
                                       ;  VL = [CL]
                           )
        ).

% vanillaConvert(VL, CL) :
% 	Convert VanillaList VL to CommaList CL

vanillaConvert(L1, L2) :- 
	is_list(L1) -> vanillaConvert1(L1, L2)
                    ;  L2 = L1.

vanillaConvert1([], true) :- !.
vanillaConvert1([G], G) :- !.
vanillaConvert1([H|T], (H, T1)) :-
	vanillaConvert1(T, T1).

%-------------------------------------------------------------
% retract_clauses/1 : retract a list of clauses

retract_clauses([]).
retract_clauses([C|Rest]) :-
	retractall(clauseI(C, _, _)),
	retract_clauses(Rest).

%-------------------------------------------------------------------
% variant_all(List)
% 	Checks whether all terms in the list are variants of each other

variant_all(L) :- member(X, L), !, map_variant(L, X).

map_variant([], _).
map_variant([H|T], X) :- variant(H, X), map_variant(T, X).


prop_match_all(L) :- member(X, L), !, map_prop_match(L, X).

map_prop_match([], _).
map_prop_match([H|T], X) :-
	prop_match(H, X, []), map_prop_match(T, X).

%-----------------------------------------------------------------

var_seteq(V1, V2) :- var_subseteq(V1, V2), var_subseteq(V2, V1).

var_subseteq([], _).
var_subseteq([X|T], L) :- var_mem(X,L), var_subseteq(T, L).

var_remove_dupl(L, L1) :- var_remove_dupl(L, [], L1).

var_remove_dupl([], X, X).
var_remove_dupl([H|T], X, Res) :-
	 ( var_mem(H, X) -> var_remove_dupl(T, X, Res)
                         ;  append(X, [H], X1),
			    var_remove_dupl(T, X1, Res)
         ).

remove_dupl(L, L1) :- remove_dupl(L, [], L1).

remove_dupl([], X, X).
remove_dupl([H|T], X, Res) :-
	 ( memberchk(H, X) -> remove_dupl(T, X, Res)
                           ;  remove_dupl(T, [H|X], Res)
         ).




set_union([], Set, Set).
set_union([H|T], Set, USet) :-
	set_union(T, Set, USet1),
	( variant_member(H, Set)
			-> USet = USet1
	                ;  USet = [H|USet1]
	).


variant_member(X, List) :- member(Y, List), variant(X, Y).

var_setunion([], U, U).
var_setunion([H|T], V, U) :-
	var_setunion(T, V, U1),
	( var_mem(H, V) -> U = U1 
	                ;  U = [H|U1]
	).


var_setintersect([], _, []).
var_setintersect([X|T], L, Res) :-
	var_setintersect(T, L, Res1),
	( var_mem(X, L) -> ( var_mem(X, Res1) -> Res = Res1
					      ;  Res = [X|Res1] )
                        ;  Res = Res1
        ).

var_setminus(V, [], V).
var_setminus(V, [H|T], R) :-
	var_remove(V, H, V1),
	var_setminus(V1, T, R).

% remove all occurrences of a particular variable

var_remove([], _, []).
var_remove([H|T], X, R) :-
	var_remove(T, X, R1),
	( H == X -> R = R1 
	         ;  R = [H|R1] 
	).

get_vars(T, List) :- get_vars1(T, L), var_remove_dupl(L, List).

get_vars1(X, [X]) :- var(X), !.
get_vars1(X, L) :- X =.. [_|Args], get_vars_list(Args, L).

get_vars_list([], []).
get_vars_list([H|T], L) :-
	get_vars1(H, L1), 
	get_vars_list(T, L2),
	append(L1, L2, L).

var_mem(X, [H|T]) :- X == H -> true ; var_mem(X, T).

vars_all([]).
vars_all([H|T]) :- var(H), vars_all(T).

var_distinct_all(L) :- var_distinct_all(L, []).

var_distinct_all([], _).
var_distinct_all([X|T], L) :-
	not(var_mem(X, L)), 
	var_distinct_all(T, [X|L]).
	

%-----------------------------------------------------------------
% Unification check without binding variables.

unifies(T1, T2) :- copy_term(T1, T3), copy_term(T2, T4), T3 = T4.

myassert(G) :- G -> true ; assert(G).


max_arg_depth(Atom, N) :-
	Atom =.. [_Pred|ArgList],
	max_term_depth(ArgList, N).

term_depth(X, 0) :- var(X), !.
term_depth(T, N) :- T =.. [_|L], max_term_depth(L, M), N is M +1.

max_term_depth([], 1).
max_term_depth([T|Rest], N) :-
	max_term_depth(Rest, X), 
	term_depth(T, D),
	(D > X -> N = D ; N = X). 


%-----------------------------------------------------------------

builtin(G) :- predicate_property(G, built_in).

execute_builtins([], []).
execute_builtins([G|Rest], Rest1) :- builtin(G), !, G, 
			             execute_builtins(Rest, Rest1).
execute_builtins([G|Rest], [G|Rest1]) :- execute_builtins(Rest, Rest1).

%-----------------------------------------------------------------------

% find_folder_clauses: find predicates in clause heads of P0

find_folder_clauses(FolderList) :-
	setof(P/N, (clause0(_, (H :- _), _), functor(H,P,N)), FPredList),
	member(FP/FN, FPredList), length(FN, L), Hd =.. [FP|L],
	findall(CN, (clause0(CN, (G :- _), _), unifies(G, Hd)), FolderList).


find_heads([], []).
find_heads([N|Rest], [H|Hrest]) :-
	clauseI(N, (H :- _), _),
	find_heads(Rest, Hrest).


remove_facts([], [], []).
remove_facts([N|Rest], CList, FList) :-
	remove_facts(Rest, CL, FL),
	( fact(N) -> CList = CL , FList = [N|FL]
		  ;  CList = [N|CL], FList = FL
        ).

fact(ClauseNum) :-
	clauseI(ClauseNum, (_ :- []), _).


% most_gen_atom(List, A)
% -- Succeeds if all atoms in List are unifiable with each other
%    and returns the most general atom in List (without binding vars).

most_gen_atom([A], A) :- !.
most_gen_atom([A|Rest], B) :-
	most_gen_atom(Rest, R),
	copy_term(A,A1), copy_term(R,R1),
	( subsumes(A1, R1)  
		-> B = A 
		;  (A1 = R1 -> B = R)
        ).


%----------------------------------------------------------------------

apply_subst([]).

apply_subst([(X = T)|Rest]) :- X = T, apply_subst(Rest).

%-------------------------------------------------------------------
%  mapping the variable occurrences of a term using a variable map


map_term(VMap, X, Y) :-
	var(X), !, ( vmap_memX((X = Y), VMap) -> true
					      ;  Y = X
                   ).

map_term(VMap, T, T1) :-
	T =.. [F|L], map_args(VMap, L, L1), T1 =.. [F|L1].

rev_map_term(VMap, Y, X) :-
	var(Y), !, ( vmap_memY((X = Y), VMap) -> true
					      ;  X = Y
                   ).

rev_map_term(VMap, T, T1) :-
	T =.. [F|L], rev_map_args(VMap, L, L1), T1 =.. [F|L1].


map_args(_VMap, [], []).
map_args(VMap, [A|Rest], [A1|Rest1]) :-
	map_term(VMap, A, A1), map_args(VMap, Rest, Rest1).

rev_map_args(_VMap, [], []).
rev_map_args(VMap, [A|Rest], [A1|Rest1]) :-
	rev_map_term(VMap, A, A1), rev_map_args(VMap, Rest, Rest1).


/*

map_term([], Term, Term).

map_term([(X = Y)|Rest], Term, New) :-
	replace_occ(X, Term, Y, Term1),
	map_term(Rest, Term1, New).


replace_occ(X, Term, Y, Term1) :-
	var(Term), !, ( Term == X -> Term1 = Y ; Term1 = Term ).

replace_occ(X, Term, Y, Term1) :-
	Term =.. [F|List], 
	replace_occ_list(X, List, Y, List1),
	Term1 =.. [F|List1].

replace_occ_list(_X, [], _Y, []).

replace_occ_list(X, [Arg|Rest], Y, [Arg1|Rest1]) :-
	replace_occ(X, Arg, Y, Arg1),
	replace_occ_list(X, Rest, Y, Rest1).

*/

%-------------------------------------------------------------------
% Do a term copy and keep track of the variable mappings


copy_term(T1, T2, Vmap) :-
	copy_term(T1, T2), get_vmap(T1, T2, Vmap0),
	var_remove_dupl(Vmap0, Vmap).

get_vmap(X, Y, [Eq]) :- var(X), var(Y), !, Eq =.. [=,X,Y].
get_vmap(X, Y, L) :- X =.. [P|Args], Y =.. [P|Args1],
  		     get_vmap_list(Args, Args1, L).

get_vmap_list([], [], []).
get_vmap_list([H|T], [H1|T1], Res) :-
	get_vmap(H, H1, L1),
	get_vmap_list(T, T1, L2), append(L1, L2, Res).


%----------------------------------------------------------------

get_clause(N, (Hd :- Bd)) :-
	clauseI(N, (H :- B), _),
	copy_term((H :- B), (Hd :- Bd)).

get_clause(N, (Hd :- Bd), Annot) :-
	clauseI(N, (H :- B), Annot),
	copy_term((H :- B), (Hd :- Bd)).

get_clause0(N, (Hd :- Bd)) :-
	clause0(N, (H :- B), _),
	copy_term((H :- B), (Hd :- Bd)).

get_clause0(N, (Hd :- Bd), Annot) :-
	clause0(N, (H :- B), Annot),
	copy_term((H :- B), (Hd :- Bd)).


% generating all the maps; a tiny heuristic here as well to 
% generate them in a prefered order.

gen_map([], _, []).
gen_map([X|L1], L2, [(X,Y)|Rest]) :-
	member(Y,L2), rotate(L2,L3), gen_map(L1, L3, Rest).

rotate([], []).
rotate([H|T], R) :- append(T, [H], R).

%-----------------------------------------------------------------

find_clauses(P/N, Cp) :-
	length(L, N), Hd =.. [P|L],
	findall(Id, clauseI(Id, (Hd :- _), _), Cp).

/*
        ( ( Cp = [], \+ predicate_property(P, built_in))
                  -> write('WARNING: Predicate '), write(P/N),
                     writeln(' has no clauses.')
                  ;  true
        ).

*/

%------------------DEAD CODE --------------------------------------
% a more accurate check : permute_clauses is being used-----------

% a conservative check to prune folded clause-set candidates

abst_foldable(Folder, Folded) :-
	abst_foldable_chk(Folder, Folded), !.

% for each folder clause, check that all the body predicates 
% are present in the body of some folded clause.

abst_foldable_chk([], _Folded).

abst_foldable_chk([C|Rest], Folded) :-
	clause0(C, (_H :- B), _), abst_atoms(B, Bd),
	all_exists_chk(Bd, Folded),
	abst_foldable_chk(Rest, Folded).

all_exists_chk(Bd, Folded) :-
	member(C, Folded), clauseI(C, (_H :- B), _),
	abst_atoms(B, B1), all_member(Bd, B1).

all_member([], _L).
all_member([A|Rest], L) :- member(A, L), all_member(Rest, L).


abst_atoms([], []).
abst_atoms([H|Rest], [H1|Rest1]) :-
	abst_atom(H, H1), abst_atoms(Rest, Rest1).

abst_atom(H, P/N) :- functor(H, P, N).

%-----------------------------------------------------------

all_equal([]).
all_equal([H|T]) :- all_equal(H, T).

all_equal(_H, []).
all_equal(H, [H|T]) :- all_equal(H, T).

%------------------------------------------------------------------
% subsumes(T1, T2, Inst)
%     check if T1 subsumes T2, and if so return the instantiation
%     without binding variables in T1, T2.  
%     Example:
%     | ?- subsumes(p(X, f(Y)), p(g(Z), f(W)), Inst).
%
%     X = _h84
%     Y = _h106
%     Z = _h144
%     W = _h170
%     Inst = [_h84 = g(_h144),_h170 = _h106]

subsumes(T1, T2, Inst) :-
	copy_term(T1, T1d, VMap1),
	copy_term(T2, T2d, VMap2), 
        subsumes(T1d, T2d),
	get_subsuming_subst(VMap1, VMap2, Inst).

get_subsuming_subst([], _VMap, []).
get_subsuming_subst([(X = T) | Rest], VMap, [(X = T1) | Rest1]) :-
	rev_map_term(VMap, T, T1), get_subsuming_subst(Rest, VMap, Rest1).


get_subst(Vmap, Inst) :- 
	get_subst(Vmap, Vmap, Inst1),
	filter_ident(Inst1, Inst).

filter_ident([], []).
filter_ident([(X = Y)|T], R) :-
	filter_ident(T, R1), 
	(X == Y -> R = R1 ; R = [(X = Y)|R1]).

get_subst([], _, []).
get_subst([(X = Y)|Rest], Vmap, [(X = T)|Irest]) :-
	get_subst_one(Y, Vmap, T),
	get_subst(Rest, Vmap, Irest).
	

get_subst_one(Y, Vmap, X) :- var(Y), !, vmap_memY((X = Y), Vmap).
get_subst_one(Y, Vmap, X) :- 
	Y =.. [P|Args],
        get_subst_one_list(Args, Vmap, Args1),	
        X =.. [P|Args1].

get_subst_one_list([], _, []).
get_subst_one_list([H|T], Vmap, [H1|T1]) :-
	get_subst_one(H, Vmap, H1),
	get_subst_one_list(T, Vmap, T1).
	
% Given X, find Y using the variable map/subst

vmap_memX((X = Y), [(Z = W)|T]) :- 
	X == Z -> Y = W ; vmap_memX((X = Y), T).
	
% Given Y, find X using the variable map/subst

vmap_memY((X = Y), [(Z = W)|T]) :- 
	Y == W -> X = Z ; vmap_memY((X = Y), T).


%--------------------------------------------------------------

% wrong_inst_local_vars(LVars, Inst, InstList)
% --  check whether Inst instantiates any local variable in 
%     LVars to a non-variable term or a non-fresh variable.

wrong_inst_local_vars([V|_Rest], I, IList) :-
	vmap_memX((V = Term), I), 
        wrong_inst_local_varchk(V, Term, IList), !.

wrong_inst_local_vars([_V|Rest], I, IList) :-
	wrong_inst_local_vars(Rest, I, IList).


wrong_inst_local_varchk(V, Term, IList) :-
	( nonvar(Term) ; 	
	  ( var(Term), non_fresh_inst(V, Term, IList)
          )
        ).
     

non_fresh_inst(V, W, [I|IRest]) :-
	non_fresh_instance(V, W, I) ;
	non_fresh_inst(V, W, IRest).

non_fresh_instance(V, W, [(X = Y) |Rest]) :-
	( W == Y, V \== X ) ; 
	non_fresh_instance(V, W, Rest).


%--------------------------------------------------------------

% sorting a list of integers

int_sort([], []).

int_sort([N|Rest], List) :-
	int_sort(Rest, List1), int_insert(N, List1, List).

int_insert(N, [], [N]).

int_insert(N, [M|Rest], List) :-
	( N > M -> int_insert(N, Rest, List1), List = [M|List1]
	        ;  List = [N,M|Rest]
        ).


%---------------------------------------------------------

% retrieveing pre-specified positions of a list 
% the list of positions is assumed to be sorted.
% call pattern (+,+,-)

retrieve_elements(List, PosList, SubList, Rest) :-
	retrieve_elements(List, PosList, 1, SubList, Rest).

retrieve_elements(L, [], _, [], L).

retrieve_elements([H|Tail], [N|NRest], Pos, SL, Rest) :-
	Pos1 is Pos + 1,
	( Pos == N -> retrieve_elements(Tail, NRest, Pos1, SL1, Rest1), 
                      SL = [H|SL1], Rest = Rest1
		   ;  retrieve_elements(Tail, [N|NRest], Pos1, SL1, Rest1),
		      SL = SL1, Rest = [H|Rest1]
        ).

%------------------------------------------------------

% my_variant/2  : have to write it only because of the bug in variant/2 

my_variant(G1, G2) :- 
	my_subsumes(G1, G2), my_subsumes(G2, G1).

my_subsumes(G1, G2) :- 
	copy_term(G1, X), copy_term(G2, Y), subsumes(X, Y).

%--------------------------------------------------------------

set_diff(L1, L2, L) :- remove_all(L2, L1, L).

non_empty_preds(PredList) :-
	setof(P/N, head_pred(P/N), PredList).

head_pred(P/N) :- clauseI(_, (Hd :- _), _), functor(Hd, P, N).

%---------------------------------------

gen_nat_list(0, []).
gen_nat_list(N, List) :-
	N > 0, M is N - 1, gen_nat_list(M, L), append(L, [N], List).


%-----------------------------------------------------------

member_pos(List, A, N) :- member_pos(List, A, 1, N).


member_pos([H|_T], H, N, N).

member_pos([_H|T], A, M, N) :-
	M1 is M+1, member_pos(T, A, M1, N).


%------------------------------------------------------

prop_approx_clause(C, (P :- B)) :-
	clauseI(C, (Hd :- Bd), _),
	functor(Hd, P, _), 
	prop_approx_list(Bd, B).

prop_approx_list([], []).

prop_approx_list([A|Rest], [P|Prest]) :-
	functor(A, P, _), prop_approx_list(Rest, Prest).


construct_pairs([], [], []).

construct_pairs([P|Rest1], [Q|Rest2], [(P,Q)|Rest3]):-
	construct_pairs(Rest1, Rest2, Rest3).